from itsimodels.core.base_models import BaseModel, ChildModel
from itsimodels.core.fields import BoolField, DictField, ForeignKeyMultiEmbeddedStr, ListField, NumberField, StringField


# Regex defined to parse the list of supported policies: "\"cef5eec4-2dcc-11eb-8ffb-0671d5072164\",\"e3ec489a-04b1-11ea-8567-021bca2da03d\",\"48a35d46-0557-11ea-9716-021bca2da03d\",\"76073f1c-303c-11eb-8ffe-0671d5072164\""
GENERIC_POLICY_RE = r'"([\w-]+?)"'


class FilterSpecification(ChildModel):
    id = StringField()

    label = StringField()

    text = StringField()

    value = ForeignKeyMultiEmbeddedStr(
        refers=None, key_regex=GENERIC_POLICY_RE)


class EventManagementState(BaseModel):
    key = StringField(required=True, alias='_key')

    title = StringField(required=True)

    arbitrary_search = StringField(alias='arbitrarySearch')

    auto_refresh = NumberField(alias='autoRefresh')

    columns_shown = ListField(object, alias='columnsShown')

    configured_tab_id = StringField(alias='configuredTabId')

    dashboard = StringField()

    description = StringField()

    earliest = StringField()

    event_deduplication = BoolField(alias='eventDeduplication')

    event_deduplication_field = StringField(alias='eventDeduplicationField')

    event_field_filter = DictField(alias='eventFieldFilter')

    fetch_limit = NumberField(alias='fetchLimit')

    filter_collection = ListField(
        FilterSpecification, alias='filterCollection')

    identifying_name = StringField()

    interactable = BoolField()

    latest = StringField()

    show_summary_dashboard = BoolField(alias='showSummaryDashboard')

    sort_direction = StringField(alias='sortDirection')

    sort_field = StringField(alias='sortField')

    view_mode = StringField(alias='viewMode')

    viewing_option = StringField(alias='viewingOption')



from itsimodels.core.base_models import BaseModel, ChildModel
from itsimodels.core.fields import (
    BoolField,
    DictField,
    ForeignKey,
    ListField,
    NumberField,
    StringField,
    TypeField
)


class FactorWeight(ChildModel):
    factor_name = StringField()

    factor_weight = NumberField()


class Rule(ChildModel):
    key = StringField(alias='_key')

    actions = ListField(object)

    activation_criteria = DictField()

    description = StringField()

    priority = NumberField()

    title = StringField()


class FieldSpecificationConfig(ChildModel):
    operator = StringField()

    field = StringField()

    value = ForeignKey('itsimodels.neap.NotableEventAggregationPolicy')


class FieldSpecification(ChildModel):
    config = TypeField(FieldSpecificationConfig)

    type = StringField()


class FieldSpecifications(ChildModel):
    items = ListField(FieldSpecification)

    condition = StringField()


class Clause(ChildModel):
    type = StringField()

    config = TypeField(FieldSpecifications)


class FilterCriteria(ChildModel):
    items = ListField(Clause)

    condition = StringField()


class NotableEventAggregationPolicy(BaseModel):
    key = StringField(required=True, alias='_key')

    title = StringField(required=True)

    breaking_criteria = DictField(object)

    description = StringField(default='')

    disabled = BoolField(default=True)

    entity_factor_enabled = BoolField(default=True)

    filter_criteria = TypeField(FilterCriteria)

    group_assignee = StringField()

    group_custom_instruction = StringField()

    group_dashboard = StringField()

    group_dashboard_context = StringField()

    group_description = StringField()

    group_instruction = StringField()

    group_severity = StringField()

    group_status = StringField()

    group_title = StringField()

    priority = StringField()

    rules = ListField(Rule)

    service_topology_enabled = BoolField(default=True)

    smart_breaking_criteria = DictField(object, alias='ace_breaking_criteria')

    smart_enabled = BoolField(default=False, alias='ace_enabled')

    smart_factor_fields = ListField(object, alias='ace_factor_fields')

    smart_factor_weights = ListField(FactorWeight, alias='ace_factor_weights')

    smart_field_analysis_end_time = NumberField(alias='ace_field_analysis_end_time')

    smart_field_analysis_start_time = NumberField(alias='ace_field_analysis_start_time')

    smart_group_assignee = StringField(alias='ace_group_assignee')

    smart_group_dashboard = StringField(alias='ace_group_dashboard')

    smart_group_dashboard_context = StringField(alias='ace_group_dashboard_context')

    smart_group_description = StringField(alias='ace_group_description')

    smart_group_instruction = StringField(alias='ace_group_instruction')

    smart_group_severity = StringField(alias='ace_group_severity')

    smart_group_status = StringField(alias='ace_group_status')

    smart_group_title = StringField(alias='ace_group_title')

    smart_split_by_field = StringField(default='', alias='ace_split_by_field')

    split_by_field = StringField(default='')

    sub_group_limit = StringField(default='')



from itsimodels.core.base_models import ImageModel
from itsimodels.core.fields import StringField


class GlassTableImage(ImageModel):
    key = StringField(required=True, alias='_key')

    data = StringField()

    name = StringField()

    type = StringField()

    # glass table images
    def get_title(self):
        return self.get_key()


from itsimodels.core.base_models import BaseModel
from itsimodels.core.compat import string_types
from itsimodels.core.fields import ForeignKey, ForeignKeyList, ListField, StringField
from itsimodels.service import EntityRule, Kpi


class ServiceTemplate(BaseModel):
    key = StringField(required=True, alias='_key')

    title = StringField(required=True)

    description = StringField(default='')

    entity_rules = ListField(EntityRule)

    kpis = ListField(Kpi)

    linked_services = ForeignKeyList('itsimodels.service.Service')

    sync_status = StringField()

    team_id = ForeignKey('itsimodels.team.Team', default='default_itsi_security_group', alias='sec_grp')

    template_tags = ListField(string_types)



from itsimodels.core.compat import string_types, text_type


class FieldValidationError(Exception):
    """Raised when a field fails to validate its value"""

    def __init__(self, value):
        """
        :param str value: the exception's error message value
        """
        super(FieldValidationError, self).__init__()

        self.value = value

    def __str__(self):
        """
        :return: the error message value
        :rtype: str
        """
        return repr(self.value)


class BaseField(object):
    """A base implementation class for a model's field definition"""

    def __init__(self, required=False, choices=None, default=None, alias=None):
        """
        :param bool required: whether the field is required
        :param tuple choices: a tuple of possible values this field can contain
        :param object default: the default value
        :param str alias: the alias name for this field
        """
        self.required = bool(required)
        self.default = default
        self.choices = choices
        self.alias = alias

    def decode(self, value):
        """
        Decodes the given value into the appropriate type

        :param value: the value
        :return: object
        """
        return value

    def validate(self, value):
        """
        Validates the given value based on the current field type

        :param value: the value to validate
        """
        if self.required and value is None:
            raise FieldValidationError('is required')


class BoolField(BaseField):
    """A field that validates boolean values"""

    def decode(self, value):
        """
        Decodes the given value into a str

        :param value: the value
        :return: bool
        """
        if value is None:
            return None

        if isinstance(value, string_types):
            value = value.lower()

        true_values = {1, '1', 'on', 't', 'true', 'yes'}

        return value in true_values

    def validate(self, value):
        """
        Validates that the given value is a boolean

        :param value: the value to validate
        """
        super(BoolField, self).validate(value)

        if value is not None and not isinstance(value, bool):
            raise FieldValidationError('should be boolean, got "{}"'.format(value))


class StringField(BaseField):
    """A field that validates string values"""

    def decode(self, value):
        """
        Decodes the given value into a str

        :param value: the value
        :return: str
        """
        return text_type(value) if value is not None else None

    def validate(self, value):
        """
        Validates that the given value is a string

        :param value: the value to validate
        """
        super(StringField, self).validate(value)

        if value is not None and not isinstance(value, string_types):
            raise FieldValidationError('should be string, got "{}"'.format(value))


class NumberField(BaseField):
    """A field that validates number values"""

    def decode(self, value):
        """
        Decodes the given value into a number

        :param value: the value
        :return: float or int
        """
        if value is None:
            return value

        if self.default is not None and isinstance(self.default, int):
            return int(value)

        return float(value)

    def validate(self, value):
        """
        Validates that the given value is a number

        :param value: the value to validate
        """
        super(NumberField, self).validate(value)

        if value is not None and not isinstance(value, (int, float)):
            raise FieldValidationError('should be number, got "{}"'.format(value))


class TypeField(BaseField):
    """A field that validates for the provided type"""

    def __init__(self, atype=None, *args, **kwargs):
        super(TypeField, self).__init__(*args, **kwargs)

        self.type = atype

    def validate(self, value):
        """
        Validates that the given value is of the given type

        :param value: the value to validate
        """
        super(TypeField, self).validate(value)

        if value is not None and not isinstance(value, self.type):
            raise FieldValidationError('should be of type="{}", got "{}"'.format(self.type, value))


class CompoundField(BaseField):
    def __init__(self, subtype=None, *args, **kwargs):
        super(CompoundField, self).__init__(*args, **kwargs)
        self.subtype = subtype


class DictField(CompoundField):
    """A field that validates dict values"""

    def __init__(self, *args, **kwargs):
        if 'default' not in kwargs:
            kwargs['default'] = {}

        super(DictField, self).__init__(*args, **kwargs)

    def validate(self, value):
        """
        Validates that the given value is a dict

        :param value: the value to validate
        """
        super(DictField, self).validate(value)

        if value is not None and not isinstance(value, dict):
            raise FieldValidationError('should be dict, got "{}"'.format(value))

        self._validate_values_type(value)

    def _validate_values_type(self, value):
        """
        Validates that the inner values have the correct type

        :param value: the iterable value to validate
        """
        if not value or self.subtype is None:
            return

        for val in list(value.values()):
            if val is not None and not isinstance(val, self.subtype):
                raise FieldValidationError(
                    'should have all {} values, got "{}"'.format(
                        self.subtype,
                        val
                    )
                )


class ListField(CompoundField):
    """A field that validates list values"""

    def __init__(self, *args, **kwargs):
        if 'default' not in kwargs:
            kwargs['default'] = []

        super(ListField, self).__init__(*args, **kwargs)

    def validate(self, value):
        """
        Validates that the given value is a list

        :param value: the value to validate
        """
        super(ListField, self).validate(value)

        if value is not None and not isinstance(value, list):
            raise FieldValidationError('should be list')

        self._validate_values_type(value)

    def _validate_values_type(self, value):
        """
        Validates that the inner values have the correct type

        :param value: the iterable value to validate
        """
        if not value or self.subtype is None:
            return

        for val in value:
            if val is not None and not isinstance(val, self.subtype):
                raise FieldValidationError(
                    'should have all {} values, got "{}"'.format(
                        self.subtype,
                        val
                    )
                )


class ForeignKey(StringField):
    """A field that represents an object id for a particular model type"""

    def __init__(self, refers, key_regex=None, *args, **kwargs):
        super(ForeignKey, self).__init__(*args, **kwargs)

        self.refers = refers
        self.key_regex = key_regex


class ForeignKeyList(ListField):
    """A field that represents a list of object ids for a particular model type"""

    def __init__(self, refers, *args, **kwargs):
        kwargs['subtype'] = string_types

        super(ForeignKeyList, self).__init__(*args, **kwargs)

        self.refers = refers


class ForeignKeyListStr(ForeignKey):
    """A field that represents a list of object ids for a particular model type
    in comma separated string format. We handle this differently during remap_keys
    from ForeignKey

    example: xxxx,yyyy,...
    """

class ForeignKeyMultiEmbeddedStr(ForeignKey):
    """A field that has multiple object ids embed in arbitrary format
    these ids could be from multiple unknown sources. Example would be search strings
    in kpi
    """



from itsimodels.core.base_models import BaseModel, ChildModel
from itsimodels.core.compat import string_types
from itsimodels.core.fields import (
    BaseField,
    BoolField,
    DictField,
    ForeignKey,
    ForeignKeyList,
    ForeignKeyMultiEmbeddedStr,
    ListField,
    NumberField,
    StringField,
    TypeField
)


KPI_SUMMARY_RE = r'`get_full_itsi_summary_kpi\((.+)\)`'


KV_STORE_KEY_RE = r'splunk-enterprise-kvstore://(.+)'

'''
Sample urls with object key embedded (one in the middle, one at the end):
Notice Id could belong to multiple types of objects. in below example, 
one is 'savedHomeView', the other is 'service'

/app/itsi/homeview?savedHomeViewId=a8579dde-dac0-11eb-8b20-0a769924fc2e&owner=nobody&view=standard&viewType=service_topology&earliest=-12h&latest=now&serviceId=bb1063f6-56c6-4ea4-8cde-b144b6b9866a&kpiId=e3fb2217f74d9b42292b28a0

/app/itsi/homeview?view=standard&viewType=service_topology&earliest=-12h&latest=now&serviceId=3f4706e8-109e-4af1-9a69-4eab5696c648
'''
EVENT_HANDLER_OPTION_URL_RE = (r'Id=([\w-]+?)(?=&|$)|'
                               r'ID=([\w-]+?)(?=&|$)')

class ValuedModel(ChildModel):

    def to_dict(self, *args, **kwargs):
        obj = super(ValuedModel, self).to_dict(*args, **kwargs)

        # Remove any null values to conform with the UI visualization library
        data = {key: value for key, value in obj.items() if
                value or (type(value) in (int, float, bool) and value is not None)}

        return data


class VisualizationOptionsModel(ValuedModel):
    """Represents an object with a dynamic field set for visualisationOptions in Glass Table JSON files."""

    def _populate(self, data, field_decoder=None):
        """
        Populates the model with values from data for each field

        :param dict data: a dict from field name to data value
        :param object field_decoder: Optional object to help parse a field value
        """
        fields = {}


        """
        Goal : The purpose of adding below logic is to support UDF visualization options : viz.* and splunk.* as mentioned in ITSI-26102

        use of variable :
        itsi_vo_compliant_fields - List of field aliases for the properties of visualizationOptions class. It is used when importbackup command is hit to package the Content Pack from backup ZIP file and ITSI objects are installed in the ITSI.
        cp_vo_model_fields - List of fields for the properties of visualizationOptions class. It is used when glass table is installed using Content Pack REST API present in ITSI
        """

        itsi_vo_compliant_fields = []
        cp_vo_model_fields = []

        for name, value in list(VisualizationOptions.__dict__.items()):
            if not isinstance(value, BaseField) or not value.alias:
                continue
            
            itsi_vo_compliant_fields.append(value.alias)
            cp_vo_model_fields.append(name)
        

        """
        This for loop is used to support splunk.* visualization options
        """
        for field_name, value in data.items():
            """
            Skipping field dict update for 2 conditions:
            (1) When the field_name is one of the field aliases present under visualizationOptions class, the model class for that field needs to be assigned as it is from VisualizationOptions class, since that class has type along with the original field name and its alias. This will be handled by model_fields method.
            (2) When the field_name is one of the field present under visualizationOptions class, as these fields are handled by model_fields method
            """
            if field_name in itsi_vo_compliant_fields or field_name in cp_vo_model_fields:
                continue
            if isinstance(value, list):
                field = ListField()
            elif isinstance(value, dict):
                field = DictField()
            elif isinstance(value, bool):
                field = BoolField()
            elif isinstance(value, (int, float)):
                field = NumberField()
            else:
                field = StringField()

            fields[field_name] = field

        """
        model_fields method will get all the fields from VisualizationOptions class and add/update the `fields` dict
        """
        fields.update(self.model_fields())

        return self._populate_with_fields(data, fields, field_decoder=field_decoder)


class DataSourceMeta(ChildModel):
    kpi_id = ForeignKey('itsimodels.service.Kpi', alias='kpiID')

    service_id = ForeignKey('itsimodels.service.Service', alias='serviceID')


class DataSourceOptions(ValuedModel):
    data = DictField()

    # NOTE:
    # glass table --> definition --> data_sources --> options --> query
    # the regex is not complete. ids may have to be manually fixed
    query = ForeignKey('itsimodels.service.Kpi', key_regex=KPI_SUMMARY_RE)

    query_parameters = DictField(alias='queryParameters')


class DataSource(ValuedModel):
    meta = TypeField(DataSourceMeta)

    name = StringField()

    options = TypeField(DataSourceOptions)

    primary = StringField()

    type = StringField()


class DefinitionInput(ChildModel):
    data_sources = DictField(alias='dataSources')

    options = DictField()

    title = StringField()

    type = StringField()


class LayoutImage(ChildModel):
    size_type = StringField(alias='sizeType')

    src = ForeignKey('itsimodels.glass_table_image.GlassTableImage', key_regex=KV_STORE_KEY_RE)

    x = NumberField()

    y = NumberField()


class LayoutOptions(ValuedModel):
    background_color = StringField(alias='backgroundColor')

    background_image = TypeField(LayoutImage, alias='backgroundImage')

    display = StringField()

    height = NumberField()

    show_title_and_description = BoolField(alias='showTitleAndDescription')

    width = NumberField()


class StructureItem(ChildModel):
    # references one of the visualizations id within the
    # glass table definition
    item = StringField()

    position = DictField()

    type = StringField()


class Layout(ChildModel):
    global_inputs = ListField(string_types, alias='globalInputs')

    options = TypeField(LayoutOptions)

    structure = ListField(StructureItem)

    type = StringField()


class VisualizationOptions(VisualizationOptionsModel):
    """ This class is providing backwards compatibility for the below mentioned properties and working as a DynamicModel class for any new property that is not mentioned below."""
    # viz.text
    # viz.rectangle
    # viz.area
    # viz.punchcard
    # viz.table
    # viz.img
    # viz.geojson.world
    # splunk.singlevalue

    area_fill_opacity = NumberField(alias='areaFillOpacity')

    axis_titlex_visibility = StringField(alias='axisTitleX__visibility')

    axis_titley_visibility = StringField(alias='axisTitleY__visibility')

    axis_labelsx_major_label_visibility = StringField(alias='axisLabelsX__majorLabelVisibility')

    axis_labelsy_major_label_visibility = StringField(alias='axisLabelsY__majorLabelVisibility')

    background_color = StringField(alias='backgroundColor')

    bubble_color = StringField(alias='bubleColor')

    color = StringField()

    color_mode = StringField(alias='colorMode')

    content = StringField()

    count = NumberField()

    fill = StringField()

    fill_color = StringField(alias='fillColor')

    font_color = StringField(alias='fontColor')

    font_family = StringField(alias='fontFamily')

    font_size = BaseField(alias='fontSize')

    font_weight = StringField(alias='fontWeight')

    foreground_color = StringField(alias='foregroundColor')

    grid_linesx_show_major_lines = BoolField(alias='gridLinesX__showMajorLines')

    grid_linesy_show_major_lines = BoolField(alias='gridLinesY__showMajorLines')

    legend_placement = StringField(alias='legend__placement')

    logical_bounds = DictField(alias='logicalBounds')

    icon = ForeignKey('itsimodels.glass_table_icon.GlassTableIcon', key_regex=KV_STORE_KEY_RE)

    major_font_size = NumberField(alias='majorFontSize')

    major_color = StringField(alias='majorColor')
    
    major_value = StringField(alias='majorValue')

    name = StringField()

    number_precision = NumberField(alias='numberPrecision')

    projection = StringField()

    preserve_aspect_ratio = BoolField(alias='preserveAspectRatio')

    rotation = NumberField()

    row_background_color_even = StringField(alias='rowBackgroundColorEven')

    row_background_color_odd = StringField(alias='rowBackgroundColorOdd')

    row_text_color_even = StringField(alias='rowTextColorEven')

    row_text_color_odd = StringField(alias='rowTextColorOdd')

    rx = NumberField()

    ry = NumberField()

    selector = StringField()

    series_colors = ListField(alias='seriesColors')

    show_bubble_labels = StringField(alias='showBubbleLabels')

    show_max_value_pulsation = BoolField(alias='showMaxValuePulsation')

    show_trend_indicator = BoolField(alias='showTrendIndicator')

    show_spark_line = BoolField(alias='showSparkline')

    show_spark_line_tooltip = BoolField(alias='showSparklineTooltip')

    show_value = BoolField(alias='showValue')

    source = StringField()

    source_bounds = DictField(alias='sourceBounds')

    spark_line_display = StringField(alias='sparklineDisplay')

    spark_line_position = StringField(alias='sparklinePosition')

    spark_line_stroke_color = StringField(alias='sparklineStrokeColor')

    spark_line_values = StringField(alias='sparklineValues')

    src = ForeignKey('itsimodels.glass_table_image.GlassTableImage', key_regex=KV_STORE_KEY_RE)

    stroke = StringField()

    stroke_width = NumberField(alias='strokeWidth')

    stroke_color = StringField(alias='strokeColor')

    stroke_highlight_color = StringField(alias='strokeHighlightColor')

    text_color = StringField(alias='textColor')
    
    trend_color = StringField(alias='trendColor')

    trend_display = StringField(alias='trendDisplay')

    trend_display_mode = StringField(alias='trendDisplayMode')

    trend_value = StringField(alias='trendValue')

    unit = StringField()

    unit_position = StringField(alias='unitPosition')

    width = NumberField()


class EventHandlerOption(ValuedModel):
    key = StringField()

    newTab = BoolField()

    type = StringField()

    # None refers because we want to check all types of objects for key
    # mapping. It is because the object key embeded could belong to different
    # types, not just one (see example above regarding EVENT_HANDLER_OPTION_URL_RE
    url = ForeignKeyMultiEmbeddedStr(refers=None, key_regex=EVENT_HANDLER_OPTION_URL_RE)


class EventHandlers(ChildModel):
    options = TypeField(EventHandlerOption)

    type = StringField()


class Visualization(ValuedModel):
    data_sources = DictField(alias='dataSources')

    encoding = DictField()

    event_handlers = ListField(EventHandlers, alias='eventHandlers')

    options = TypeField(VisualizationOptions)

    # references visualization ids within glass table def
    primary = StringField()

    title = StringField()

    type = StringField()

    context = DictField()


class Definition(ChildModel):
    data_sources = DictField(DataSource, alias='dataSources')

    defaults = DictField()

    description = StringField()

    inputs = DictField(DefinitionInput)

    layout = TypeField(Layout)

    title = StringField()

    visualizations = DictField(Visualization)


class GlassTable(BaseModel):
    key = StringField(required=True, alias='_key')

    title = StringField(required=True)

    definition = TypeField(Definition)

    description = StringField(default='')

    gt_version = StringField(default='beta')

    # interactable = BoolField()

    latest = StringField(default='now')

    latest_label = StringField(default='Now')

    swap_service_ids = ForeignKeyList('itsimodels.service.Service')

    selected_swap_service_id = ForeignKey('itsimodels.service.Service')

    template_selected_service_id = ForeignKey('itsimodels.service.Service', alias='templateSelectedServiceId')

    template_swappable_service_ids = ForeignKeyList('itsimodels.service.Service', alias='templateSwappableServiceIds')

